from enum import Enum
from pathlib import Path
from warnings import warn

from .util import logger, get_iesopt_module_attr
from .julia.util import jl_symbol, recursive_convert_py2jl, jl_safe_seval
from .results import Results


class ModelStatus(Enum):
    """Status of an :py:class:`iesopt.Model`."""

    EMPTY = "empty"
    GENERATED = "generated"
    FAILED_GENERATE = "failed_generate"
    FAILED_OPTIMIZE = "failed_optimize"
    OPTIMAL = "optimal"
    OPTIMAL_LOCALLY = "local_optimum"
    INFEASIBLE = "infeasible"
    INFEASIBLE_OR_UNBOUNDED = "infeasible_unbounded"
    OTHER = "other"


class Model:
    """An IESopt model, based on an :jl:module:`IESopt.jl <IESopt.IESopt>` core model."""

    def __init__(self, filename: str | Path, **kwargs) -> None:
        self._filename = filename
        self._kwargs = recursive_convert_py2jl(kwargs)

        self._model = None
        self._verbosity = kwargs.get("verbosity", True)

        self._status = ModelStatus.EMPTY
        self._status_details = None

        self._results = None

        self._IESopt = get_iesopt_module_attr("IESopt")
        self._JuMP = get_iesopt_module_attr("JuMP")
        self._jump_value = get_iesopt_module_attr("jump_value")
        self._jump_dual = get_iesopt_module_attr("jump_dual")

        if "virtual_files" in self._kwargs:
            # Ensure proper type of this entry, since it is enforced by `IESopt.parse!`.
            _dict_typed = self._IESopt.seval("Dict{String, DataFrames.DataFrame}")
            self._kwargs["virtual_files"] = _dict_typed(self._kwargs["virtual_files"])

    def __repr__(self) -> str:
        if self._model is None:
            return "An IESopt model (not yet generated)"

        n_var = self._JuMP.num_variables(self.core)
        n_con = self._JuMP.num_constraints(self.core, count_variable_in_set_constraints=False)
        solver = self._JuMP.solver_name(self.core)
        termination_status = self._JuMP.termination_status(self.core)

        return (
            f"An IESopt model:"
            f"\n\tname: {self.data.input.config['general']['name']['model']}"
            f"\n\tsolver: {solver}"
            f"\n\t"
            f"\n\t{n_var} variables, {n_con} constraints"
            f"\n\tstatus: {termination_status}"
        )

    @property
    def core(self):
        """Access the core `JuMP` model that is used internally."""
        if self._model is None:
            raise Exception("Model was not properly set up; call `generate` first")
        return self._model

    @property
    def data(self):
        """Access the IESopt data object of the model.

        This is deprecated; use `model.internal` instead (similar to the Julia usage `IESopt.internal(model)`).
        """
        warn(
            "Using `model.data` is deprecated; use `model.internal` instead (similar to the Julia usage `IESopt.internal(model)`)",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.internal

    @property
    def internal(self):
        """Access the IESopt data object of the model."""
        return self._IESopt.internal(self.core)

    @property
    def status(self) -> ModelStatus:
        """Get the current status of this model. See `ModelStatus` for possible values."""
        return self._status

    @property
    def objective_value(self) -> float:
        """Get the objective value of the model. Only available if the model was solved beforehand."""
        if self._status == ModelStatus.OPTIMAL_LOCALLY:
            logger.warning("Model is only locally optimal; objective value may not be accurate")
        elif self._status != ModelStatus.OPTIMAL:
            raise Exception("Model is not optimal; no objective value available")
        return self._JuMP.objective_value(self.core)

    def generate(self) -> None:
        """Generate a IESopt model from the attached top-level YAML config."""
        try:
            self._model = self._IESopt.generate_b(str(self._filename), **self._kwargs)
            self._status = ModelStatus.GENERATED
        except Exception as e:
            self._status = ModelStatus.FAILED_GENERATE
            logger.error(f"Exception during `generate`: {e}")
            try:
                logger.error(f"Current debugging info: {self.data.debug}")
            except Exception as e:
                logger.error("Failed to extract debugging info")

    def write_to_file(self, filename=None, *, format: str = "automatic") -> str:
        """Write the model to a file.

        Consult the Julia version of this function, [IESopt.write_to_file](https://ait-energy.github.io/iesopt/pages/manual/julia/index.html#write-to-file)
        for more information. This will automatically invoke `generate` if the model has not been generated yet.

        Arguments:
            filename (Optional[str]): The filename to write to. If `None`, the path and name are automatically
                                      determined by IESopt.

        Keyword Arguments:
            format (str): The format to write the file in. If `automatic`, the format is determined based on the
                          extension of the filename. Otherwise, it used as input to [`JuMP.write_to_file`](https://jump.dev/JuMP.jl/stable/api/JuMP/#write_to_file),
                          by converting to uppercase and prefixing it with `MOI.FileFormats.FORMAT_`. Writing to, e.g.,
                          an LP file can be done by setting `format="lp"` or `format="LP"`.

        Returns:
            str: The filename (including path) of the written file.

        Examples:
            ..  code-block:: python
                :caption: Writing a model to a problem file.

                import iesopt

                cfg = iesopt.make_example("01_basic_single_node", dst_dir="opt")

                # Model will be automatically generated when calling `write_to_file`:
                model = iesopt.Model(cfg)
                model.write_to_file()

                # It also works with already optimized models:
                model = iesopt.run(cfg)
                model.write_to_file("opt/out/my_problem.LP")

                # And supports different formats:
                target = model.write_to_file("opt/out/my_problem.foo", format="mof")
                print(target)
        """
        if self._status == ModelStatus.EMPTY:
            self.generate()

        try:
            if filename is None:
                return self._IESopt.write_to_file(self.core)
            else:
                format = jl_safe_seval(f"JuMP.MOI.FileFormats.FORMAT_{format.upper()}")
                return self._IESopt.write_to_file(self.core, str(filename), format=format)
        except Exception as e:
            logger.error(f"Error while writing model to file: {e}")
            return ""

    def optimize(self) -> None:
        """Optimize the model."""
        try:
            self._IESopt.optimize_b(self.core)

            if self._JuMP.is_solved_and_feasible(self.core, allow_local=False):
                self._status = ModelStatus.OPTIMAL
                self._results = Results(model=self)
            elif self._JuMP.is_solved_and_feasible(self.core, allow_local=True):
                self._status = ModelStatus.OPTIMAL_LOCALLY
                self._results = Results(model=self)
            else:
                _term_status = self._JuMP.termination_status(self.core)
                if str(_term_status) == "INFEASIBLE":
                    self._status = ModelStatus.INFEASIBLE
                    logger.error(
                        "The model seems to be infeasible; refer to `model.compute_iis()` and its documentation if you want to know more about the source of the infeasibility."
                    )
                elif str(_term_status) == "INFEASIBLE_OR_UNBOUNDED":
                    self._status = ModelStatus.INFEASIBLE_OR_UNBOUNDED
                else:
                    self._status = ModelStatus.OTHER
                    self._status_details = _term_status
        except Exception as e:
            self._status = ModelStatus.FAILED_OPTIMIZE
            logger.error(f"Exception during `optimize`: {e}")
            try:
                logger.error(f"Current debugging info: {self.data.debug}")
            except Exception as e:
                logger.error("Failed to extract debugging info")

    def compute_iis(self, filename=None) -> None:
        """Compute and print the Irreducible Infeasible Set (IIS) of the model, or optionally write it to a file.

        Note that this requires a solver that supports IIS computation.

        Arguments:
            filename (Optional[str | Path]): The filename to write the IIS to. If `None`, the IIS is only printed to the
                                             console.

        Examples:
            ..  code-block:: python
                :caption: Computing the IIS of a model.

                import iesopt

                model = iesopt.run("infeasible_model.iesopt.yaml")
                model.compute_iis()

                # or (arbitrary filename/extension):
                model.compute_iis(filename="my_problem.txt")
        """
        try:
            if filename is None:
                self._IESopt.compute_IIS(self.core)
            else:
                self._IESopt.compute_IIS(self.core, filename=str(filename))
        except Exception as e:
            logger.error("Error while computing IIS: `%s`" % str(e.args[0]))

    @property
    def results(self) -> Results:
        """Get the results of the model."""
        if self._results is None:
            raise Exception("No results available; have you successfully called `optimize`?")
        return self._results

    def extract_result(self, component: str, field: str, mode: str = "value"):
        """Manually extract a specific result from the model."""
        try:
            c = self._IESopt.get_component(self.core, component)
        except Exception:
            raise Exception(f"Exception during `extract_result({component}, {field}, mode={mode})`")

        f = None
        for fieldtype in ["var", "exp", "con", "obj"]:
            try:
                t = getattr(c, fieldtype)
                f = getattr(t, field)
                break
            except Exception:
                pass

        if f is None:
            raise Exception(f"Field `{field}` not found in component `{component}`")

        try:
            if mode == "value":
                return self._jump_values(f)
            elif mode == "dual":
                return self._jump_duals(f)
            else:
                raise Exception(f"Mode `{mode}` not supported, use `value` or `dual`")
        except Exception:
            raise Exception(f"Error during extraction of result `{field}` from component `{component}`")

    def get_component(self, component: str):
        """Get a core component based on its full name."""
        try:
            return self._IESopt.get_component(self.core, component)
        except Exception:
            raise Exception(f"Error while retrieving component `{component}` from model")

    def get_components(self, tagged=None):
        """
        Get all components of the model, possibly filtered by (a) tag(s).

        Arguments:
            tagged (str or list of str): The tag(s) to filter the components by, can be `None` (default) to get all components.
        """
        if tagged is None:
            return self._IESopt.get_components(self.core)

        return self._IESopt.get_components(self.core, recursive_convert_py2jl(tagged))

    def get_variable(self, component: str, variable: str):
        """Get a specific variable from a core component."""
        raise DeprecationWarning(
            f"`get_variable(...)` is deprecated; you can instead access the variable directly using "
            f"`model.get_component('{component}').var.{variable}`. Conversion to Python lists, if not done "
            f"automatically, can be done using `list(...)`."
        )

    def get_constraint(self, component: str, constraint: str):
        """Get a specific constraint from a core component."""
        raise DeprecationWarning(
            f"`get_constraint(...)` is deprecated; you can instead access the constraint directly using "
            f"`model.get_component('{component}').con.{constraint}`. Conversion to Python lists, if not done "
            f"automatically, can be done using `list(...)`."
        )

    def nvar(self, var: str):
        """Extract a named variable, from `model`.

        If your variable is called `:myvar`, and you would access it in Julia using `model[:myvar]`, you can call
        `model.nvar("myvar")`.
        """
        try:
            return self.core[jl_symbol(var)]
        except Exception:
            raise Exception(f"Error while retrieving variable `{var}` from model")

    @staticmethod
    def _to_pylist(obj):
        # if isinstance(obj, jl.VectorValue):
        #    return jl.PythonCall.pylist(obj)
        return obj



import sys
from pathlib import Path
import importlib.metadata

from sphinxawesome_theme.postprocess import Icons


# -- Setup -------------------------------------------------------------------

DOCS = Path(__file__).parent
sys.path.insert(0, str((DOCS / "_extensions").resolve()))
sys.path.insert(0, str((DOCS / ".." / "src").resolve()))
sys.path.insert(0, str((DOCS / "..").resolve()))

CREATE_DYNAMIC_DOCSTRINGS = [
    # ("test", "getfield", "some_foo(...)")
    # and then use:
    # ```{include} ../../../dynamic/out/test.md
    # ```
]
if len(CREATE_DYNAMIC_DOCSTRINGS) > 0:
    import docs.dynamic.create as ddcreate

    ddcreate.create_md(CREATE_DYNAMIC_DOCSTRINGS)

import docs.dynamic.core as ddcore  # noqa: E402

ddcore._dyn_core_create_md("Connection")
ddcore._dyn_core_create_md("Decision")
ddcore._dyn_core_create_md("Node")
ddcore._dyn_core_create_md("Profile")
ddcore._dyn_core_create_md("Unit")

import docs.dynamic.julia  # noqa: E402, F401

# -- General -----------------------------------------------------------------

project = "IESopt"
copyright = "AIT Austrian Institute of Technology GmbH"
author = "S. Strömer (@sstroemer), D. Schwabeneder (@daschw)"
version = str(importlib.metadata.version("iesopt"))
release = version

extensions = [
    # "myst_parser",
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx_togglebutton",
    "julia_domain",
    "sphinx.ext.mathjax",
]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]
source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}
nb_custom_formats = {
    ".md": ["jupytext.reads", {"fmt": "myst"}],
}

intersphinx_mapping = {
    "IESopt": ("https://ait-energy.github.io/IESopt.jl/dev", None),
    "python": ("https://docs.python.org/3", None),
    "pandas": ("http://pandas.pydata.org/pandas-docs/dev", None),
}
intersphinx_disabled_reftypes = [""]

nb_execution_mode = "off"

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinxawesome_theme"
html_static_path = ["_static"]
html_css_files = ["custom.css"]

html_permalinks_icon = Icons.permalinks_icon

pygments_style = "default"
pygments_style_dark = "lightbulb"

html_theme_options = {
    "show_breadcrumbs": True,
    "show_scrolltop": True,
    "awesome_external_links": True,
    "awesome_headerlinks": True,
    "show_prev_next": False,
    "main_nav_links": {
        "Home": "index",
        "GitHub": "https://github.com/ait-energy/iesopt",
        "IESopt.jl": "https://github.com/ait-energy/IESopt.jl",
    },
}

html_context = {
    "default_mode": "light",
}

html_favicon = "_static/favicon.ico"

# -- MyST & autodoc configuration --------------------------------------------

myst_enable_extensions = ["dollarmath", "amsmath", "colon_fence", "fieldlist"]
myst_heading_anchors = 3
autodoc_packages = ["../src/iesopt"]
suppress_warnings = ["autosectionlabel.*"]
autosection_label_prefix_document = True



import iesopt


class TestPythonJuliaConversion:
    def test_kwargs_dict(self, tmp_path):
        # Refer to: https://github.com/ait-energy/IESopt.jl/issues/66

        configs = [
            {"a": 1, "b": 2.0, "c": "string"},
            {"a": 1, "c": "string"} | {"b": 2.0},
            {"b": 2.0} | {"a": 1, "c": "string"},
        ]

        for config in configs:
            model = iesopt.Model("", config=config)
            assert type(model._kwargs["config"]["a"]) is int



from pathlib import Path

from .util import logger, get_iesopt_module_attr
from .model import Model, ModelStatus


def run(filename: str | Path, **kwargs) -> Model:
    r"""
    Generate and optimize an IESopt model.

    Results can be accessed (after a successful optimization) using `model.results`.

    Arguments:
        filename : str
            Path to the IESopt model file to load.

    Keyword Arguments:
        **kwargs: Additional keyword arguments to pass to the `Model` constructor.

    Returns:
        The generated and optimized IESopt model.

    Example:
        ..  code-block:: python
            :caption: Run an IESopt model

            import iesopt
            iesopt.run("opt/config.iesopt.yaml")
    """
    model = Model(filename, **kwargs)
    model.generate()

    if model.status == ModelStatus.GENERATED:
        model.optimize()
    else:
        logger.error("Model could not be generated; skipping optimization.")

    return model


def examples() -> list[str]:
    """
    Return a list of all available examples.

    Returns:
        List of available examples. This contains the names of the examples, not the full filenames (so, e.g.,
        `"some_example"` instead of `"some_example.iesopt.yaml"`).
    """
    import os

    julia = get_iesopt_module_attr("julia")
    folder = Path(str(julia.IESopt.Assets.get_path("examples")))
    return sorted([fn.split(".iesopt.yaml")[0] for fn in os.listdir(folder) if fn.endswith(".iesopt.yaml")])


def make_example(example: str, dst_dir: str | Path = "./", dst_name: str | None = None) -> Path:
    """
    Generate a local copy of a specific example.

    A list of examples, and their exact names, can be obtained using :py:func:`iesopt.examples()`.

    Arguments:
        example : str
            Name of the example to generate.
        dst_dir : Optional[str]
            Directory to generate the example in, defaults to `"./"`.
        dst_name: Optional[str]
            Name of the generated example file (without the ".iesopt.yaml" extension), e.g.,
            `"config"`, will create `dst_dir/config.iesopt.yaml`. Will default to the original name of the example.

    Returns:
        Path to the generated example file.
    """
    import shutil
    import os
    import stat

    julia = get_iesopt_module_attr("julia")
    folder = Path(str(julia.IESopt.Assets.get_path("examples")))

    filename = folder / f"{example}.iesopt.yaml"
    datafolder = folder / "files"
    target_filename = Path(dst_dir) / ((example if dst_name is None else dst_name) + ".iesopt.yaml")
    target_datafolder = Path(dst_dir) / "files"

    # Check if data folder already exists.
    if not os.path.exists(target_datafolder):
        logger.info("Data folder for examples does not exist; creating it, and copying contents")
        shutil.copytree(datafolder, target_datafolder)
    else:
        logger.info("Data folder for examples already exists; NOT copying ANY contents")

    # Copy the config file.
    logger.info("Creating example ('%s') at: '%s'" % (example, target_filename))
    shutil.copy(filename, target_filename)

    # Make everything editable.
    logger.info(
        "Set write permissions for example ('%s'), and data folder ('%s')" % (target_filename, target_datafolder)
    )
    os.chmod(target_filename, os.stat(target_filename).st_mode & ~stat.S_IWRITE | stat.S_IWUSR)
    for dirpath, _, filenames in os.walk(target_datafolder):
        os.chmod(dirpath, os.stat(dirpath).st_mode & ~stat.S_IWRITE | stat.S_IWUSR)
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            os.chmod(filepath, os.stat(filepath).st_mode & ~stat.S_IWRITE | stat.S_IWUSR)

    return target_filename



import iesopt
from pathlib import Path


class TestExamples:
    def write_to_file(self, tmp_path):
        model = iesopt.Model(iesopt.make_example("01_basic_single_node", dst_dir=tmp_path))
        assert Path(model.write_to_file()).exists()
        assert Path(model.write_to_file(tmp_path / "my_problem.LP")).exists()
        assert Path(model.write_to_file(tmp_path / "my_problem.foo", format="mof")).exists()



import warnings
import re
import pandas as pd
import numpy as np
from pydantic import validate_call

from .util import get_iesopt_module_attr
from .julia import jl_isa


class ddict(dict):
    """Wrapper to enable dot.notation access to dictionary attributes."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class JuliaCallPyConvertReadWrapper:
    def __init__(self, obj):
        import juliacall

        self._juliacall = juliacall
        self._obj = obj

    def keys(self):
        if isinstance(self._obj, dict | self._juliacall.DictValue):
            return dict(self._obj).keys()
        raise TypeError(f"Object of type '{type(self._obj)}' inside JuliaCallPyConvertReadWrapper has no keys")

    def __len__(self):
        if isinstance(self._obj, dict | list | tuple | self._juliacall.DictValue | self._juliacall.VectorValue):
            return len(self._obj)
        raise TypeError(f"Object of type '{type(self._obj)}' inside JuliaCallPyConvertReadWrapper has no length")

    def _rewrap_return(self, value):
        if isinstance(value, self._juliacall.ArrayValue):
            return value.to_numpy(copy=False)
        if isinstance(value, float | int | str | bool | type(None)):
            return value

        if isinstance(value, self._juliacall.DictValue):
            return JuliaCallPyConvertReadWrapper(value)

        # Some other/unknown type, rewrap it.
        return JuliaCallPyConvertReadWrapper(value)

    def __getattr__(self, name: str):
        return self._rewrap_return(getattr(self._obj, name))

    def __getitem__(self, key):
        return self._rewrap_return(self._obj.get(key))


class Results:
    _valid_attrs = ["attributes", "model", "custom", "input", "info", "snapshots", "components"]

    @validate_call
    def __init__(self, *, file: str = None, model=None):
        """
        Create a new `Results` object, either from a file or from an IESopt model. Make sure to pass either a file or a
        model explicitly using a keyword argument.

        Keyword Arguments
            file : Optional[str]
                Path to the results file to load, by default None
            model : Optional[IESopt.Model]
                IESopt model to extract results from, by default None
        """
        self._attributes = None
        self._model = None
        self._snapshots = None
        self._components = None
        self._custom = None
        self._input = None
        self._info = None
        self._source = None

        self._cache = None

        self._IESopt = get_iesopt_module_attr("IESopt")
        self._julia = get_iesopt_module_attr("julia")

        if (int(file is None) + int(model is None)) != 1:
            raise Exception("Either file or model must be set, not both and not none of them.")
        elif file is not None:
            if not isinstance(file, str):
                raise Exception("File must be a `str`, did you try passing a model without `model=your_model`?")
            self._from_file(file)
        elif model is not None:
            self._from_model(model)

        # Allow dot access to attributes of "model" results, to align with access to the models structure.
        self._model = ddict(self._model)

    @property
    def components(self):
        return JuliaCallPyConvertReadWrapper(self._model["components"])

    @property
    def objectives(self):
        return JuliaCallPyConvertReadWrapper(self._model["objectives"])

    @property
    def customs(self):
        return JuliaCallPyConvertReadWrapper(self._model["customs"])

    @validate_call
    def get(
        self,
        result: str,
        component: str,
        fieldtype: str,
        field: str,
        *,
        mode: str = "primal",
        build_cache: bool = False,
    ):
        if result not in ["component", "objective", "custom"]:
            raise ValueError(f"`result` must be 'component', 'objective', or 'custom', got '{result}'")

        if result in ["objective", "custom"]:
            # TODO: Implement and remove this
            raise NotImplementedError("Accessing objectives and custom results is not yet supported using `get(...)`")

        if mode == "dual":
            field = field + "__dual"
        elif mode != "primal":
            raise ValueError(f"`mode` must be 'primal' or 'dual', got '{mode}'")

        if build_cache:
            self._build_cache()

        if self._has_cache():
            t = (component, fieldtype, field)
            if t in self._cache:
                return self._cache[t]
            raise ValueError(f"Failed to access result '{fieldtype}.{field}' in component '{component}'")

        return self._get_safe(component, fieldtype, field, mode)

    @validate_call
    def query_available_results(self, component: str, mode: str = "both"):
        """
        Query the available results for a specific component, optionally filtered by `mode`.

        Arguments:
            component : str
                Component to query results for
            mode : Optional[str]
                Mode to query results for, either "both", "primal", or "dual", by default "both"
        """
        regex = re.compile(component)

        if not self._has_cache():
            self._build_cache()

        if mode == "both":
            return [
                (it[1], (x := it[2].split("__"))[0], "primal" if len(x) == 1 else "dual")
                for it in self._cache
                if re.search(regex, it[0])
            ]
        elif mode == "primal":
            return [(it[1], it[2]) for it in self._cache if re.search(regex, it[0]) and "__dual" not in it[2]]
        elif mode == "dual":
            return [
                (it[1], it[2].split("__")[0]) for it in self._cache if re.search(regex, it[0]) and "__dual" in it[2]
            ]
        else:
            raise ValueError("Invalid mode, must be either 'both', 'primal', or 'dual'")

    @validate_call
    def entries(self, field: str = None):
        """
        Get all available entries for a specific field, or all fields if `field` is `None`.

        Arguments:
            field : Optional[str]
                Field to get entries for, by default None which returns all fields
        """
        if field is None:
            return [it for it in Results._valid_attrs if getattr(self, f"_{it}") is not None]
        return sorted(getattr(self, field).keys())

    @validate_call
    def to_dict(self, filter=None, field_types=None, build_cache: bool = True) -> dict:
        """
        Extract results from the `Results` object and return them as a dictionary.

        Arguments:
            filter : Optional[Callable]
                Filter function to apply to the results, by default None, must take three arguments: (c, t, f) where `c` is
                the component's name, `t` is the field type ("var", "exp", "con", or "obj"), and `f` is the field name.
            field_types : Optional[list[str]]
                Field types to extract, by default None, which extracts all field types.
            build_cache : Optional[bool]
                Whether to build a cache of the results, by default True.

        Returns:
            Dictionary of results, with keys as tuples of (component, fieldtype, field) and values as the result.
        """
        if build_cache:
            self._build_cache()

        if field_types is None:
            field_types = ["var", "exp", "con", "obj", "res"]

        entries = {}
        if self._cache is None:
            for c in self._model["components"].keys():
                for t in field_types:
                    container = getattr(self._model["components"][c], t)
                    for f in [str(it) for it in self._julia.Main.keys(container)]:
                        if (filter is None) or filter(c, t, f):
                            entries[(c, t, f)] = Results._safe_convert(getattr(container, f))
        else:
            for c, t, f in self._cache.keys():
                if (t in field_types) and ((filter is None) or filter(c, t, f)):
                    entries[(c, t, f)] = self._cache[(c, t, f)]

        return entries

    @validate_call
    def to_pandas(self, filter=None, field_types=None, orientation: str = "long", build_cache: bool = True):
        """
        Extract results from the `Results` object and return them as :py:class:`pandas.DataFrame` or :py:class:`pandas.Series` (depending on the
        shape of the included results).

        Arguments:
            filter : Optional[Callable]
                Filter function to apply to the results, by default None, must take three arguments: (c, t, f) where `c` is
                the component's name, `t` is the field type ("var", "exp", "con", or "obj"), and `f` is the field name
            field_types : Optional[list[str]]
                Field types to extract, by default None, which extracts all field types
            orientation : Optional[str]
                Orientation of the resulting DataFrame, either "wide" or "long", by default "long"
            build_cache : Optional[bool]
                Whether to build a cache of the results, by default True.

        Returns:
            DataFrame or Series of results, depending on the orientation and shape of the results.
        """
        _dict = self.to_dict(filter, field_types, build_cache)

        # Check if we can cast to `pd.Series` (no matter the orientation).
        _result_types = set(type(it) for it in _dict.values())
        if len(_result_types) == 1:
            _rt = _result_types.pop()
            if _rt in [int, float]:
                return pd.Series(_dict)

            if _rt == np.ndarray:
                if len(_dict) == 1:
                    k, v = next(iter(_dict.items()))

                    if isinstance(v, int | float):
                        return pd.Series(v, index=k)

                    value_shape = v.shape
                    if len(value_shape) == 1:
                        if value_shape[0] == len(self._snapshots):
                            return pd.Series(v, index=self._snapshots, name=k)

            if orientation == "wide":
                return pd.DataFrame(_dict, index=self._snapshots)

        if orientation == "wide":
            warnings.warn(
                "Results must be of the same shape (= length) to create a DataFrame, with `orientation='wide'`. "
                "Falling back to `orientation='long'`. This warning can be prevented by calling `results.to_pandas("
                "orientation='long') instead, or by applying filters (using `filter` and/or `field_types`) that only "
                "select results of the same shape (example: objectives are always scalars, while most variables - "
                "except Decisions - are vector valued, indexed over all snapshots)."
            )
            warnings.warn(f"Got the following result types: {_result_types}")
            orientation = "long"

        if orientation == "long":
            n_snapshots = len(self._snapshots)
            _data = {c: [] for c in ["snapshot", "component", "fieldtype", "field", "value", "mode"]}

            for (c, t, f), v in _dict.items():
                m = "primal"
                if f.endswith("__dual"):
                    f = f[:-6]
                    m = "dual"

                if isinstance(v, (int, float)):
                    _data["snapshot"].append(None)
                    _data["component"].append(c)
                    _data["fieldtype"].append(t)
                    _data["field"].append(f)
                    _data["value"].append(v)
                    _data["mode"].append(m)
                else:
                    _data["snapshot"].extend(self._snapshots)
                    _data["component"].extend([c] * n_snapshots)
                    _data["fieldtype"].extend([t] * n_snapshots)
                    _data["field"].extend([f] * n_snapshots)
                    _data["value"].extend(v)
                    _data["mode"].extend([m] * n_snapshots)

            try:
                return pd.DataFrame(_data)
            except Exception:
                warnings.warn(
                    "Failed to create DataFrame. This is mostlikely due to non aligned result shapes. A "
                    "common cause are custom results, registered inside an addon, that have a different "
                    "temporal resolution than the main model results. Consider filtering out such results."
                )
                return None

        raise ValueError(f"`orientation` can be 'wide' or 'long', got '{orientation}'.")

    @validate_call
    def overview(self, component: str, *, temporal: bool, mode: str = "both"):
        regex = re.compile(component)

        if not self._has_cache():
            self._build_cache()

        if mode == "both":
            ret = {
                (k[0], k[1], (x := k[2].split("__"))[0], "primal" if len(x) == 1 else "dual"): v
                for (k, v) in self._cache.items()
                if re.search(regex, k[0]) and isinstance(v, int | float) != temporal
            }
        elif mode == "primal":
            ret = {
                (k[0], k[1], k[2], "primal"): v
                for (k, v) in self._cache.items()
                if re.search(regex, k[0]) and "__dual" not in k[2] and isinstance(v, int | float) != temporal
            }
        elif mode == "dual":
            ret = {
                (k[0], k[1], k[2].split("__")[0], "dual"): v
                for (k, v) in self._cache.items()
                if re.search(regex, k[0]) and "__dual" in k[2] and isinstance(v, int | float) != temporal
            }
        else:
            raise ValueError("Invalid mode, must be either 'both', 'primal', or 'dual'")

        if len(ret) == 0:
            return None

        if temporal:
            return pd.DataFrame(ret, index=self._snapshots).sort_index(axis=1)
        else:
            return pd.Series(ret)

    # def __getattr__(self, attr: str):
    #     if attr not in Results._valid_attrs:
    #         raise Exception(f"Attribute '{attr}' is not accessible, pick one of {Results._valid_attrs}")
    #     if not hasattr(self, f"_{attr}"):
    #         raise Exception(f"`Results` object has no attribute '{attr}'")
    #     if getattr(self, f"_{attr}") is None:
    #         raise Exception(f"Attribute '{attr}' not properly loaded, consider checking `results.entries()` first")
    #     return getattr(self, f"_{attr}")

    def _build_cache(self):
        if self._cache is None:
            self._cache = self.to_dict(build_cache=False)

    def _has_cache(self):
        return self._cache is not None

    def _get_safe(self, component: str, fieldtype: str, field: str, mode: str = "primal"):
        if mode not in ["primal", "dual"]:
            raise ValueError(f"`mode` must be 'primal' or 'dual', got '{mode}'.")

        try:
            components = self._model["components"]
        except Exception:
            raise ValueError("Failed to access `components` in model results")
        try:
            all_results = components[component]
        except Exception:
            raise ValueError(f"Failed to access component '{component}' in model results")
        try:
            container = getattr(all_results, fieldtype)
        except Exception:
            raise ValueError(f"Failed to access results for `fieldtype` '{fieldtype}' in component '{component}'")
        try:
            result = getattr(container, field)
        except Exception:
            raise ValueError(f"Failed to access result '{field}' in component '{component}'")

        return Results._safe_convert(result)

    def _from_file(self, file: str):
        results = self._IESopt.ResultsJLD2.load_results(file)

        result_entries = ddict({})
        for key in results.keys():
            _current = result_entries
            _keys = key.split("/")
            for i in range(len(_keys)):
                if i == len(_keys) - 1:
                    _current[_keys[i]] = Results._safe_convert(results[key])
                else:
                    _current[_keys[i]] = _current.get(_keys[i], ddict({}))
                    _current = _current[_keys[i]]

        for key in result_entries.keys():
            if hasattr(self, f"_{key}"):
                if getattr(self, f"_{key}") is None:
                    setattr(self, f"_{key}", result_entries[key])
                else:
                    raise Exception(f"Duplicate result entry: {key}")
            else:
                raise Exception(f"Unknown result entry: {key}")

        snapshots = results["model/snapshots"]
        self._snapshots = [snapshots[t].name for t in sorted(snapshots.keys())]
        self._components = sorted(results["model/components"].keys())
        self._source = file

    def _from_model(self, model):
        self._source = "an IESopt model"
        self._model = {
            "components": model.internal.results.components,
            "objectives": model.internal.results.objectives,
            "custom": model.internal.results.customs,
        }
        self._snapshots = [model.internal.model.snapshots[t].name for t in model.internal.model.T]
        self._components = sorted(model.internal.results.components.keys())

    def __repr__(self) -> str:
        _sep = "', '"
        return (
            "An IESopt result object:"
            + f"\n\tsource: {self._source}"
            + "\n\tattributes: "
            + (
                f"'{self.entries('attributes')[0]}', ..., '{self.entries('attributes')[-1]}'"
                if self._attributes is not None
                else "none"
            )
            + f"\n\tmodel results: {'yes' if self._model is not None else 'no'}"
            + "\n\tcustom results: "
            + (f"{len(self._custom)}" if self._custom is not None and len(self._custom) > 0 else "no")
            + "\n\tinput fields: "
            + (f"'{_sep.join(self._input)}'" if self._input is not None else "none")
            + "\n\tinfo fields: "
            + (f"'{_sep.join(self._info)}'" if self._info is not None else "none")
        )

    @classmethod
    def _safe_convert(cls, value):
        # TODO: ensure recursive conversion of all multi-element types
        if jl_isa(value, "AbstractVector"):
            return value.to_numpy(copy=False)
        if jl_isa(value, "AbstractDict"):
            return dict(value)
        if jl_isa(value, "AbstractSet"):
            return set(value)
        return value



from dotenv import dotenv_values


def _strtobool(val: str):
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    if val in ("n", "no", "f", "false", "off", "0"):
        return 0
    raise ValueError("Invalid truth value %r" % (val,))


class Config:
    DEFAULTS = {
        "IESOPT_JULIA": "1.11.5",
        "IESOPT_CORE": "2.6.4",
        "IESOPT_JUMP": "1.25.0",
        "IESOPT_SOLVER_HIGHS": "1.15.0",
        "IESOPT_MULTITHREADED": "no",  # yes, no
        "IESOPT_OPTIMIZATION": "latency",  # rapid, latency, normal, performance
    }
    _config = None

    @classmethod
    def init(cls):
        if cls._config is not None:
            return

        cls._config = {
            (k[7:] if k.startswith("IESOPT_PKG_") else k[7:].lower()): v
            for (k, v) in {
                **cls.DEFAULTS,
                **dotenv_values(),
            }.items()
            if k.startswith("IESOPT_")
        }

    @classmethod
    def get(cls, key: str):
        value = cls._config[key]
        if key in ["multithreaded"]:
            return _strtobool(value)
        return value

    @classmethod
    def find(cls, prefix: str):
        for k in cls._config.keys():
            if k.startswith(prefix):
                yield k


Config.init()



import iesopt
from pathlib import Path


# Make sure we are using the latest version.
# This is necessary to pull in changes that may be present when an `.env` file with
# ```
# IESOPT_CORE = "https://github.com/username/IESopt.jl#feature-branch-name"
# ```
# is present in the root of the repository.
# This takes a bit of time even if nothing changes, but it's rather quick in that case.
iesopt.julia.seval('import Pkg; Pkg.update("IESopt")')


DYNAMIC = Path(__file__).parent

DYN_CORE_HEADER_NOTE = """```{note}
This section of the documentation is auto-generated from the code of the Julia-based core model. Refer to [IESopt.jl](https://github.com/ait-energy/IESopt.jl) for any further details (which may require some familiarity with Julia).

**If you spot incorrect math-mode rendering**, or similar issues, please [file an issue](https://github.com/ait-energy/iesopt/issues), since rendering documentation from Julia to Python is not the easiest task.
```

## Overview

"""


# def _md_table_to_dict(md: str):
#     lines = md.strip().splitlines()
#     headers = [header.strip() for header in lines[0].split("|")[1:-1]]
#     data = []

#     for row in lines[2:]:  # skip header and divider lines
#         values = [value.strip() for value in row.split("|")[1:-1]]
#         data.append(dict(zip(headers, values)))

#     return data


# def _trf_overview(doc: str):
#     # doc = doc.replace('!!! details "Basic Examples"', ":::{admonition} \"Basic Examples\"\n:class: dropdown\n")
#     doc = doc.replace('!!! details "Basic Examples"', "## Basic Examples\n\n")
#     doc = re.sub(r"^ {0,4}", "", doc, flags=re.MULTILINE)
#     return doc + "\n"


# def _trf_parameters(doc: str):
#     ret = ""
#     data = _md_table_to_dict(doc)

#     for item in data:
#         ret += f"### {item.pop('Name').replace('`', '')}\n\n"
#         ret += f"{item.pop('Description')}\n\n"
#         for k, v in item.items():
#             ret += f":{k}: {v}\n"

#     return ret


# def _add_details(info: str):
#     ret = ""
#     for entry in info.split("!!! details ")[1:]:
#         entry = re.sub(r"^ {0,4}", "", entry, flags=re.MULTILINE)
#         entry = re.sub(r'!!! tip "([^"]+)"', r":::{admonition} \1\n    :class: dropdown", entry)
#         entry = entry.replace("!!! tip", ":::{tip}")
#         entry = re.sub(r'!!! info "([^"]+)"', r":::{admonition} \1\n    :class: dropdown", entry)
#         entry = entry.replace("!!! info", ":::{hint}")
#         # TODO: add other admonition conversions
#         lines = entry.splitlines()
#         title = lines[0].replace('"', "")
#         ret += f"\n\n#### {title}" + "\n\n"
#         is_in_adm = False
#         for line in lines[1:]:
#             if len(line) == 0:
#                 continue

#             if is_in_adm and line[0] != "\n" and line[0] != " ":
#                 is_in_adm = False
#                 ret += ":::\n\n"

#             if is_in_adm:
#                 ret += line[4:] + "\n"
#             else:
#                 ret += line + "\n"
#             if line.startswith(":::"):
#                 is_in_adm = True
#         if is_in_adm:
#             ret += ":::\n\n"
#     return ret


# def _trf_details(doc: str):
#     tmp = doc.strip().split("## Variables")[1]

#     tmp = tmp.split("## Expressions")
#     details_var = tmp[0].strip()

#     tmp = tmp[1].split("## Constraints")
#     details_exp = tmp[0].strip()

#     tmp = tmp[1].split("## Objectives")
#     details_con = tmp[0].strip()
#     details_obj = tmp[1].strip()

#     ret = "### Variables\n\n"
#     ret += _add_details(details_var)

#     ret += "### Expressions\n\n"
#     ret += _add_details(details_exp)

#     ret += "### Constraints\n\n"
#     ret += _add_details(details_con)

#     ret += "### Objectives\n\n"
#     ret += _add_details(details_obj)

#     return ret


def _getdd(cc: str):
    dt = iesopt.julia.seval(f"IESopt.{cc}")
    return iesopt.IESopt._get_dynamic_documentation(dt)


def _dyn_core_create_md(cc: str):
    docs = _getdd(cc)

    main_docstr = docs["docstr_main"]
    final_main_docstr = ""
    if '!!! details "Basic Examples"' in main_docstr:
        within_example = False
        for line in main_docstr.splitlines():
            if not within_example:
                if line == '!!! details "Basic Examples"':
                    final_main_docstr += ":::{admonition} **Basic Examples**\n:class: dropdown\n\n"
                    within_example = True
                else:
                    final_main_docstr += line + "\n"
            else:
                final_main_docstr += line[4:] + "\n"
        final_main_docstr = final_main_docstr + ":::"
    else:
        final_main_docstr = main_docstr

    with open(DYNAMIC / ".." / "pages" / "manual" / "yaml" / "core" / f"{cc.lower()}.md", "w") as f:
        f.write(f"# {cc}\n\n")
        f.write(DYN_CORE_HEADER_NOTE)
        f.write(f"{final_main_docstr}\n\n")

        f.write("## Parameters\n\n")
        for field in docs["fields_all"]:
            if field not in docs["docstr_fields"]:
                continue

            f.write(f"### `{field}`\n\n")
            f.write(f"{docs['docstr_fields'][field]['description']}\n\n")

            for p in ["mandatory", "default", "values", "unit"]:
                val = docs["docstr_fields"][field][p]
                if p == "default":
                    val = val.strip("`")
                    val = f"${val}$"
                f.write(f"- **{p}:** {val}\n")

            f.write("\n\n")

        f.write("## Detailed reference\n\n\n")

        for detail in ["Expressions", "Variables", "Constraints", "Objectives"]:
            f.write(f"### {detail}\n\n")

            funcs = docs["functions"][detail[0:3].lower()]
            for func in funcs:
                info = docs["docstr_functions"][func]
                f.write(f"#### `{info['name']}`\n\n")

                f.write(
                    f":::{{admonition}} How to access this {info['type_long']}?\n:class: dropdown\n\n"
                    f"```julia\n"
                    f"# Using Julia (`IESopt.jl`):\n"
                    f"import IESopt\n\n"
                    f"model = IESopt.run(...)  # assuming this is your model\n"
                    f'IESopt.get_component(model, "your_{info["component"]}").{info["type_short"]}.{info["name"]}\n'
                    f"```\n\n"
                    f"```python\n"
                    f"# Using Python (`iesopt`):\n"
                    f"import iesopt\n\n"
                    f"model = iesopt.run(...)  # assuming this is your model\n"
                    f'model.get_component("your_{info["component"]}").{info["type_short"]}.{info["name"]}\n'
                    f"```\n\n"
                    f":::\n\n"
                )

                f.write(
                    f"Full implementation and all details: [`{info['component']}/{info['type_short']}_{info['name']} @ IESopt.jl`](https://github.com/{info['code_path']})\n\n"
                )
                ds = info["docstr"]
                is_in_math_mode = False
                is_in_admonition = False
                for line in ds.splitlines():
                    if len(line) == 0:
                        if is_in_math_mode:
                            f.write("\n")
                        elif not is_in_admonition:
                            # This is necessary to create a new line in markdown, because Sphinx swallows these ...
                            f.write("> \n> &nbsp;\n> \n")
                        continue

                    line = line.lstrip(">")

                    if line.lstrip("> ").startswith("$$"):
                        # Math mode from Julia Markdown.
                        if is_in_math_mode:
                            f.write("$$\n")
                        else:
                            f.write("$$\n")
                        is_in_math_mode = not is_in_math_mode
                        continue
                    elif line.startswith("!!!"):
                        # Admonition from Julia Markdown.
                        assert not is_in_admonition  # currently can't handle nested admonitions
                        is_in_admonition = not is_in_admonition
                        res = line.split(maxsplit=2)
                        if len(res) == 2:
                            f.write(f":::{{{res[1]}}}\n\n")
                        elif len(res) == 3:
                            _stripped = res[2].strip('"')
                            f.write(f":::{{admonition}} **{_stripped}**\n:class: {res[1]}\n\n")
                        continue

                    if is_in_math_mode:
                        f.write(f"{line.lstrip('> ')}\n")
                    elif is_in_admonition:
                        if line.startswith("    "):
                            f.write(f"{line.lstrip()}\n")
                        else:
                            f.write(":::\n")
                            f.write(f"> {line.lstrip()}\n")
                            is_in_admonition = not is_in_admonition
                    else:
                        f.write(f"> {line.lstrip()}\n")

                if is_in_admonition:
                    f.write(":::\n")
                # ds = str(info["docstr"]).replace("\n", "\n> ")  ## TODO: remove str
                # f.write(f"> {ds}\n\n")



import pandas as pd
from ..util import logger, get_iesopt_module_attr


def jl_safe_seval(code: str):
    """Safely evaluate Julia code."""
    try:
        return get_iesopt_module_attr("julia").seval(code)
    except Exception as e:
        logger.error("Exception while trying to execute Julia code: `%s`" % code)
        logger.error("Exception: `%s`" % str(e))
        logger.error("Exception details: `%s`" % getattr(e, "exception", "missing"))

    return None


def jl_import(module: str):
    """Import a Julia module."""
    logger.info("    %s" % module)
    jl_safe_seval(f"import {module}")


def jl_symbol(string: str):
    """Create a Julia `Symbol` from `string`.

    This function should be called as `iesopt.Symbol`, but can also be called as `iesopt.jl_symbol`.

    Arguments:
        string (str): The string to convert to a Julia `Symbol`.

    Returns:
        The Julia `Symbol` object.

    Examples:
        ..  code-block:: python
            :caption: Directly import and use the `Symbol` alias

            from iesopt import Symbol
            Symbol(":iesopt")               # => `Julia: :iesopt`

    """
    return get_iesopt_module_attr("julia").Symbol(string)


def jl_isa(obj, julia_type: str):
    julia = get_iesopt_module_attr("julia")

    if julia_type == "AbstractVector":
        return julia.isa(obj, julia.AbstractVector)

    if julia_type == "AbstractSet":
        return julia.isa(obj, julia.AbstractSet)

    if julia_type == "AbstractDict":
        return julia.isa(obj, julia.AbstractDict)

    return julia.seval(f"(x) -> (x isa {julia_type})")(obj)


def jl_docs(obj: str, module: str = "IESopt"):
    """Get the documentation string of a Julia object inside the IESopt module."""
    julia_docstr = jl_safe_seval(f"@doc {module}.{obj}")
    if julia_docstr is None:
        return "Missing documentation string."
    return "".join(el for el in julia_docstr.text)


def recursive_convert_py2jl(item):
    if isinstance(item, dict):
        julia = get_iesopt_module_attr("julia")
        convert = get_iesopt_module_attr("juliacall").convert
        return convert(julia.seval("Dict{String, Any}"), {k: recursive_convert_py2jl(v) for (k, v) in item.items()})
    elif isinstance(item, list):
        julia = get_iesopt_module_attr("julia")
        convert = get_iesopt_module_attr("juliacall").convert
        return convert(julia.seval("Vector{Any}"), [recursive_convert_py2jl(v) for v in item])
    elif isinstance(item, pd.DataFrame):
        julia = get_iesopt_module_attr("julia")
        return julia.IESopt.DataFrames.DataFrame(item)

    return item



import iesopt

import pandas as pd
import hashlib


def _check_example_model(name: str, dst_dir, checks: dict, mode: str = "check"):
    config_file = iesopt.make_example(name, dst_dir=str(dst_dir), dst_name="config")
    model = iesopt.run(config_file, config={"general.verbosity.core": "error"})

    if mode == "check":
        assert str(model.status) == checks["status"]
    else:
        assert mode == "print"
        print(str(model.status))

    obj = model.objective_value

    if checks and (checks["objective_value"] == "moa"):
        if mode == "check":
            assert obj is not None
            assert not isinstance(obj, float)
        else:
            print(obj)
    else:
        if mode == "check":
            assert (obj - checks["objective_value"]) < 1e-4
        else:
            print(obj)

    if (mode == "print") or checks["hash"]:
        res = model.results.to_pandas().round(4)  # rounding to prevent solver version changes to impact tests too much
        hash = hashlib.sha1(pd.util.hash_pandas_object(res, index=True).values).hexdigest()[0:10]

        if mode == "check":
            assert hash == checks["hash"]
        else:
            print(hash)

    return model


class TestExamples:
    def test_example_list(self):
        example_list = iesopt.examples()
        assert isinstance(example_list, list)
        assert len(example_list) > 0
        assert all(isinstance(it, str) for it in example_list)

    def test_make_example(self, tmp_path):
        example_list = iesopt.examples()
        example = example_list[0]

        config_file = iesopt.make_example(example, dst_dir=tmp_path)
        assert config_file.exists()
        assert (tmp_path / f"{example}.iesopt.yaml").exists()
        assert (tmp_path / "files").exists()

        iesopt.make_example(example, dst_dir=tmp_path / "examples", dst_name="config")
        assert (tmp_path / "examples" / "config.iesopt.yaml").exists()
        assert (tmp_path / "examples" / "files").exists()

    def test_example_01(self, tmp_path):
        # _check_example_model("01_basic_single_node", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "01_basic_single_node",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 525.0, "hash": "935baad73e"},
        )

    def test_example_02(self, tmp_path):
        # _check_example_model("02_advanced_single_node", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "02_advanced_single_node",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 1506.75, "hash": "24a1836199"},
        )

    def test_example_03(self, tmp_path):
        # _check_example_model("03_basic_two_nodes", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "03_basic_two_nodes",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 1225.0, "hash": "fb3c6d290e"},
        )

    def test_example_04(self, tmp_path):
        # _check_example_model("04_soft_constraints", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "04_soft_constraints",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 1000002975.0, "hash": "7dc6e24399"},
        )

    def test_example_05(self, tmp_path):
        # _check_example_model("05_basic_two_nodes_1y", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "05_basic_two_nodes_1y",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 667437.75, "hash": "b785df3592"},
        )

    def test_example_06(self, tmp_path):
        # _check_example_model("06_recursion_h2", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "06_recursion_h2",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 18790.8, "hash": "f44f7448f9"},
        )

    def test_example_07(self, tmp_path):
        # _check_example_model("07_csv_filestorage", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "07_csv_filestorage",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 667437.75, "hash": "b785df3592"},
        )

    def test_example_08(self, tmp_path):
        # _check_example_model("08_basic_investment", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "08_basic_investment",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 2015.642857142858, "hash": "9926bf435a"},
        )

    def test_example_09(self, tmp_path):
        # _check_example_model("09_csv_only", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "09_csv_only",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 667437.75, "hash": "b785df3592"},
        )

    def test_example_10(self, tmp_path):
        # _check_example_model("10_basic_load_shedding", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "10_basic_load_shedding",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 26083.94444444445, "hash": "1c2b722d2d"},
        )

    def test_example_11(self, tmp_path):
        # _check_example_model("11_basic_unit_commitment", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "11_basic_unit_commitment",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 1570.0, "hash": "5e486c8d95"},
        )

    def test_example_12(self, tmp_path):
        # _check_example_model("12_incremental_efficiency", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "12_incremental_efficiency",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 3570.0, "hash": "e34e2ce586"},
        )

    def test_example_15(self, tmp_path):
        # _check_example_model("15_varying_efficiency", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "15_varying_efficiency",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 2131435.7798025194, "hash": "fd5c3ba13f"},
        )

    def test_example_16(self, tmp_path):
        # _check_example_model("16_noncore_components", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "16_noncore_components",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 4372.20640901404, "hash": "05b67f7585"},
        )

    def test_example_17(self, tmp_path):
        # _check_example_model("17_varying_connection_capacity", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "17_varying_connection_capacity",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 300.0, "hash": "a641898bfe"},
        )

    def test_example_18(self, tmp_path):
        # _check_example_model("18_addons", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "18_addons",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 51.0, "hash": None},
        )

    def test_example_20(self, tmp_path):
        # _check_example_model("20_chp", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "20_chp",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 16687.5, "hash": "76020dfd6b"},
        )

    def test_example_22(self, tmp_path):
        # _check_example_model("22_snapshot_weights", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "22_snapshot_weights",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 215.0, "hash": "6247daa427"},
        )

    def test_example_23(self, tmp_path):
        # _check_example_model("23_snapshots_from_csv", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "23_snapshots_from_csv",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 215.0, "hash": "6a86c37a6b"},
        )

    def test_example_25(self, tmp_path):
        # _check_example_model("25_global_parameters", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "25_global_parameters",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 50.0, "hash": "e6cbeb2547"},
        )

    def test_example_27(self, tmp_path):
        # _check_example_model("27_piecewise_linear_costs", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "27_piecewise_linear_costs",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 450.0, "hash": "adb8ef62b3"},
        )

    def test_example_29(self, tmp_path):
        # _check_example_model("29_advanced_unit_commitment", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "29_advanced_unit_commitment",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 7000.0, "hash": "b826bfcee4"},
        )

    def test_example_31(self, tmp_path):
        # _check_example_model("31_exclusive_operation", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "31_exclusive_operation",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": -10.0, "hash": "da99231d6b"},
        )

    # def test_example_35(self, tmp_path):
    #     _check_example_model(
    #         "35_fixed_costs",
    #         tmp_path,
    #         {
    #             "status": "ModelStatus.OPTIMAL",
    #             "objective_value": 91539936.26783474,
    #             "hash": "4b3ba12112f436acb43fda8cc446d12bc8e8742d4ddfcaa6c27dfa74a8ee655c",
    #         },
    #     )

    def test_example_37(self, tmp_path):
        # _check_example_model("37_certificates", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "37_certificates",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 44376.75, "hash": None},
        )

    def test_example_41(self, tmp_path):
        # _check_example_model("41_multiobjective_epsilon", "/tmp/iesopt-tmp/", None, "print")
        model = _check_example_model(
            "41_multiobjective_epsilon",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": "moa", "hash": "7b1af169ea"},
        )

        assert len(model.objective_value) == 2
        assert abs(model.objective_value[0] - 2625.0) < 1e-4
        assert abs(model.objective_value[1] - 7.5) < 1e-4

    def test_example_43(self, tmp_path):
        # _check_example_model("43_multiobjective_hierarchical", "/tmp/iesopt-tmp/", None, "print")
        model = _check_example_model(
            "43_multiobjective_hierarchical",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": "moa", "hash": "8e87105785"},
        )

        assert len(model.objective_value) == 3
        assert abs(model.objective_value[0] - 3281.25) < 1e-4
        assert abs(model.objective_value[1] - 7.459780388151175) < 1e-4
        assert abs(model.objective_value[2] - 52.298901940755876) < 1e-4

    def test_example_44(self, tmp_path):
        # _check_example_model("44_lossy_connections", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "44_lossy_connections",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 1233.75, "hash": "c4573c804d"},
        )

    def test_example_45(self, tmp_path):
        # _check_example_model("45_result_extraction", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "45_result_extraction",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 1427.9999999999998, "hash": "78edd55a2d"},
        )

    def test_example_46(self, tmp_path):
        # _check_example_model("46_constants_in_objective", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "46_constants_in_objective",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 3015.642857142858, "hash": "9926bf435a"},
        )

    def test_example_47(self, tmp_path):
        # _check_example_model("47_disable_components", "/tmp/iesopt-tmp/", None, "print")
        _check_example_model(
            "47_disable_components",
            tmp_path,
            {"status": "ModelStatus.OPTIMAL", "objective_value": 550.0, "hash": "44b1d5e46a"},
        )



import iesopt
from pathlib import Path


# Make sure we are using the latest version.
# This is necessary to pull in changes that may be present when an `.env` file with
# ```
# IESOPT_CORE = "https://github.com/username/IESopt.jl#feature-branch-name"
# ```
# is present in the root of the repository.
# This takes a bit of time even if nothing changes, but it's rather quick in that case.
iesopt.julia.seval('import Pkg; Pkg.update("IESopt")')


DYNAMIC = Path(__file__).parent

DYN_CORE_HEADER_NOTE = """```{note}
This section of the documentation is auto-generated from the code of the Julia-based core model. Refer to [IESopt.jl](https://github.com/ait-energy/IESopt.jl) for any further details (which may require some familiarity with Julia).

**If you spot incorrect math-mode rendering**, or similar issues, please [file an issue](https://github.com/ait-energy/iesopt/issues), since rendering documentation from Julia to Python is not the easiest task.
```

## Overview

"""


def doc_julia_module(module):
    entries = {"functions": [], "macros": [], "types": [], "modules": {}}

    for name in iesopt.julia.names(module):
        obj = iesopt.julia.getfield(module, name)

        if obj == module:
            continue

        if iesopt.jl_isa(obj, "Module"):
            if str(name) in ["IESU"]:
                # Just an abbreviation.
                continue

            # Create a subpage for this.
            entries["modules"][str(name)] = doc_julia_module(obj)
            # TODO: this does not work since stuff is not exported internally...
        elif iesopt.jl_isa(obj, "Function"):
            if str(name).startswith("@"):
                entries["macros"].append((str(name), obj))
            else:
                entries["functions"].append((str(name), obj))
        elif iesopt.jl_isa(obj, "Type"):
            # A struct, check if we should exclude it.
            if str(name) in ["Connection", "Decision", "Node", "Profile", "Unit"]:
                continue
            entries["types"].append((str(name), obj))
        else:
            print("ERROR - Unknown object type:", name, iesopt.julia.typeof(obj))

    return entries


entries = doc_julia_module(iesopt.IESopt)


def split_docstr(ds: str):
    if ds[1] != " ":
        return ("", ds)

    head = ""
    content = ""

    in_head = True
    for line in ds.split("\n"):
        if (not line.startswith("    ")) and (len(line) > 1):
            in_head = False
        if in_head:
            head += line[4:] + "\n"
        else:
            content += line + "\n"

    return (head, content)


def escape_md(md: str):
    ret = ""
    within_code = False
    for line in md.splitlines(True):
        if line.startswith("```"):
            within_code = not within_code

        if within_code:
            ret += line
            continue
        else:
            line = line.strip(" ")

        if line.startswith("#"):
            line = line.strip("#")
            line = line.strip(" \n")
            line = f"_**{line}**_  \n"

        ret += line

    return ret


with open(DYNAMIC / ".." / "pages" / "manual" / "julia" / "index.md", "w") as f:
    with open(DYNAMIC / ".." / "pages" / "manual" / "julia" / "index.md_template", "r") as ft:
        f.write(ft.read())

    for header in ["Types", "Macros", "Functions"]:
        f.write(f"### {header}\n\n")
        for item in entries[header.lower()]:
            if item[0].startswith("_"):
                # Skip "private" stuff that may "exported" for internal reasons.
                continue
            f.write(f"#### `{item[0]}`\n\n")

            docstr = split_docstr(iesopt.jl_docs(item[0], "IESopt"))
            if docstr[0] != "":
                f.write(f"```julia\n{docstr[0]}\n```\n\n")
            if docstr[1] != "":
                f.write(f"{escape_md(docstr[1])}\n\n")

            f.write("---\n\n")


for module in entries["modules"]:
    with open(DYNAMIC / ".." / "pages" / "manual" / "julia" / f"{module.lower()}.md", "w") as f:
        f.write(f"# {module}\n\n")
        f.write(DYN_CORE_HEADER_NOTE)

        f.write(split_docstr(iesopt.jl_docs(module))[1])
        f.write("\n\n")

        f.write("## API Reference\n\n")

        for header in ["Types", "Macros", "Functions"]:
            f.write(f"### {header}\n\n")
            for item in entries["modules"][module][header.lower()]:
                if item[0].startswith("_"):
                    # Skip "private" stuff that may "exported" for internal reasons.
                    continue
                f.write(f"#### `{item[0]}`\n\n")

                docstr = split_docstr(iesopt.jl_docs(item[0], f"IESopt.{module}"))
                if docstr[0] != "":
                    f.write(f"```julia\n{docstr[0]}\n```\n\n")
                if docstr[1] != "":
                    f.write(f"{escape_md(docstr[1])}\n\n")

                f.write("---\n\n")



import sys


def set_iesopt_module_attr(attr: str, value):
    """Set an internal attribute of the `iesopt` module."""
    sys.modules["iesopt"].__dict__[f"_!_attr_{attr}"] = value


def get_iesopt_module_attr(attr: str):
    """Get an internal attribute of the `iesopt` module."""
    return sys.modules["iesopt"].__dict__[f"_!_attr_{attr}"]



# See: DocumenterInterLinks.jl
# > https://github.com/JuliaDocs/DocumenterInterLinks.jl/
# >     "sphinx-to-documenter-links"
from sphinx.domains import Domain, ObjType
from sphinx.roles import XRefRole


class JuliaDomain(Domain):
    """A minimal Julia language domain."""

    name = "jl"
    label = "Julia"
    object_types = {
        # name => (localized name, *roles)
        "macro": ObjType("macro", "macro", "obj"),
        "keyword": ObjType("keyword", "keyword", "obj"),
        "function": ObjType("function", "func", "function", "obj"),
        "method": ObjType("method", "meth", "method", "obj"),
        "type": ObjType("type", "type", "obj"),
        "module": ObjType("module", "mod", "module", "obj"),
        "constant": ObjType("constant", "const", "constant", "obj"),
    }

    roles = {
        "macro": XRefRole(),
        "keyword": XRefRole(),
        "function": XRefRole(fix_parens=True),
        "func": XRefRole(fix_parens=True),
        "method": XRefRole(),
        "meth": XRefRole(),
        "type": XRefRole(),
        "module": XRefRole(),
        "mod": XRefRole(),
        "constant": XRefRole(),
        "const": XRefRole(),
        "obj": XRefRole(),
    }

    def resolve_any_xref(self, *args):
        return []


def setup(app):
    app.add_domain(JuliaDomain)



import iesopt
from pathlib import Path


DYNAMIC = Path(__file__).parent


def create_md(items):
    for it in items:
        with open(DYNAMIC / "out" / f"{it[0]}.md", "w") as f:
            f.write("`````{attribute} " + it[2] + "\n")
            f.write(iesopt.get_jl_docstr(it[1]))
            f.write("\n`````")



import os
import importlib.metadata
from pathlib import Path

# Set version.
__version__ = importlib.metadata.version("iesopt")

# Set juliapkg target path.
__target__ = Path().cwd() / ".iesopt"

# Set sysimage path.
__sysimage__ = __target__ / ("sysimage_v" + __version__.replace(".", "-") + ".so")

# =======================================================================
# Setup "module globals" that will be overwritten internally.
julia = None
"""
The Julia object, if setup was successful.

..  code-block:: python
    :caption: Usage

    import iesopt
    iesopt.julia.println("Hello, from Julia's world!")
"""

Symbol = None

IESopt = None
"""IESopt.jl module from Julia."""

JuMP = None
"""JuMP.jl module from Julia."""


def jump_value(item):
    """Calls `JuMP.value.(item)`."""
    pass


def jump_dual(item):
    """Calls `JuMP.dual.(item)`."""
    pass


def jump_reduced_cost(item):
    """Calls `JuMP.reduced_cost.(item)`."""
    pass


def jump_shadow_price(item):
    """Calls `JuMP.shadow_price.(item)`."""
    pass


def get_jl_docstr(obj: str):
    """Get the documentation string of a Julia object inside the IESopt module."""
    pass


# =======================================================================

# Setup Julia.
from .julia import initialize as _initialize_everything  # noqa: E402
from .util import get_iesopt_module_attr as _get_iesopt_module_attr  # noqa: E402

julia = _initialize_everything(__target__, __sysimage__)

# Export everything.
from iesopt.model import Model as Model, ModelStatus as ModelStatus  # noqa: E402
from iesopt.results import Results as Results  # noqa: E402
from iesopt.iesopt import run as run, examples as examples, make_example as make_example  # noqa: E402

from .julia.util import jl_symbol as jl_symbol, jl_docs  # noqa: E402
from .julia import jl_isa as jl_isa  # noqa: E402

Symbol = jl_symbol

if not os.getenv("IESOPT_DOCS_NOEXEC"):
    IESopt = _get_iesopt_module_attr("IESopt")
    JuMP = _get_iesopt_module_attr("JuMP")

    jump_value = _get_iesopt_module_attr("jump_value")
    jump_dual = _get_iesopt_module_attr("jump_dual")
    jump_reduced_cost = _get_iesopt_module_attr("jump_reduced_cost")
    jump_shadow_price = _get_iesopt_module_attr("jump_shadow_price")
    get_jl_docstr = jl_docs


def create_sysimage():
    """Create a sysimage containing IESopt.jl and important dependencies."""
    import juliacall
    import juliapkg

    julia.Pkg.add("PackageCompiler")
    julia.seval("import PackageCompiler")

    __target__.mkdir(exist_ok=True)

    juliapkg.add("Pkg", "44cfe95a-1eb2-52ea-b672-e2afdf69b78f", target=str(__target__ / "juliapkg.json"))
    juliapkg.resolve()

    julia.PackageCompiler.create_sysimage(
        juliacall.convert(julia.Vector, ["IESopt", "JuMP", "HiGHS", "Pkg"]), sysimage_path=str(__sysimage__)
    )

    print("Successfully created sysimage; launch your script/code again to use it")
    exit(0)



import os
import sys
from pathlib import Path
import ssl

from ..util import logger
from .util import jl_import
from ..config import Config


def lookup_package(name: str):
    lookup = {
        "jump": ("JuMP", "4076af6c-e467-56ae-b986-b466b2749572"),
        "highs": ("HiGHS", "87dc4568-4c63-4d18-b0c0-bb2238e4078b"),
        "iesopt": ("IESopt", "ed3f0a38-8ad9-4cf8-877e-929e8d190fe9"),
        "gurobi": ("Gurobi", "2e9cd046-0924-5485-92f1-d5272153d98b"),
        "cplex": ("CPLEX", "a076750e-1247-5638-91d2-ce28b192dca0"),
        "ipopt": ("Ipopt", "b6b21f68-93f8-5de0-b562-5493be1d77c9"),
        "pythoncall": ("PythonCall", "6099a3de-0909-46bc-b1f4-468b9a2dfc0d"),
    }

    if name in lookup:
        return lookup[name]

    raise Exception(f"Failed to lookup Julia package '{name}'; please report this issue")


def add_package(f_add, name: str, config: str, target: str):
    if "github" in config:
        if "#" in config:
            url, rev = config.split("#")
            f_add(*lookup_package(name), url=url, rev=rev, target=target)
        else:
            f_add(*lookup_package(name), url=config, target=target)
    else:
        f_add(*lookup_package(name), version="=" + config, target=target)


def setup_julia(target: Path, sysimage: Path):
    target.mkdir(exist_ok=True)
    target_fullpath = str(target.resolve())
    logger.info(f"    Target for juliapkg: '{target_fullpath}'")

    logger.info("Checking Julia environment")

    if "juliacall" in sys.modules:
        logger.error(
            "It seems juliacall, and thus Julia, is already loaded; this may lead to unexpected behavior and prevents"
            " proper setup based on the internal configs. If you are sure this is not an issue, you can safely ignore"
            " this message, but we cannot guarantee anything to work as expected."
        )

    if Path("juliapkg.json").exists():
        raise Exception("Found `juliapkg.json` file; remove it to prevent potential conflicts")

    if (target / "juliapkg.json").exists():
        (target / "juliapkg.json").unlink()

    # Check for local SSL certificate file, that can interfere with Julia setup.
    _ssl = None
    if "SSL_CERT_FILE" in os.environ:
        logger.debug("Detected local `SSL_CERT_FILE`; disabling it during Julia setup")
        _ssl = os.environ.pop("SSL_CERT_FILE")

    ssl._create_default_https_context = ssl._create_unverified_context
    logger.info("Disabling SSL verification to prevent problems; this may be unsafe")

    # Set `JULIA_SSL_CA_ROOTS_PATH` to prevent various SSL related issues (with Julia setup; LibGit2; etc.).
    if "JULIA_SSL_CA_ROOTS_PATH" in os.environ:
        if os.environ["JULIA_SSL_CA_ROOTS_PATH"] != "":
            logger.info(
                "Overwriting the env. variable `JULIA_SSL_CA_ROOTS_PATH` (current: `%s`) to prevent SSL issues during the Julia setup"
                % str(os.environ["JULIA_SSL_CA_ROOTS_PATH"])
            )
            os.environ["JULIA_SSL_CA_ROOTS_PATH"] = ""
    else:
        logger.debug('Setting `JULIA_SSL_CA_ROOTS_PATH = ""` to prevent SSL issues during the Julia setup')
        os.environ["JULIA_SSL_CA_ROOTS_PATH"] = ""

    # Setup Julia (checking if it "looks" valid).
    sys.path.insert(0, target_fullpath)
    import juliapkg

    # Set Julia version.
    juliapkg.require_julia(f"={Config.get('julia')}", target=target_fullpath)
    # add_package(juliapkg.add, "pythoncall", "0.9.23", target)

    # Set versions of "core" packages.
    add_package(juliapkg.add, "jump", Config.get("jump"), target=target_fullpath)
    add_package(juliapkg.add, "iesopt", Config.get("core"), target=target_fullpath)

    # Set versions of "solver" packages.
    for entry in Config.find("solver_"):
        name = entry[7:]
        add_package(juliapkg.add, name, Config.get(entry), target=target_fullpath)

    if not juliapkg.resolve(dry_run=True):
        logger.warning("The Julia environment is dirty and needs to be resolved, which can take some time")
        if not juliapkg.resolve(dry_run=False):
            raise Exception("Julia setup is not valid")

    logger.info("Julia environment ready, loading Julia")

    os.environ["JULIA_PKG_PRESERVE_TIERED_INSTALLED"] = "true"
    os.environ["PYTHON_JULIACALL_STARTUP_FILE"] = "no"
    os.environ["PYTHON_JULIACALL_AUTOLOAD_IPYTHON_EXTENSION"] = "no"

    if Config.get("multithreaded"):
        os.environ["PYTHON_JULIACALL_THREADS"] = "auto"
        os.environ["PYTHON_JULIACALL_HANDLE_SIGNALS"] = "yes"
    else:
        os.environ["PYTHON_JULIACALL_THREADS"] = "1"
        os.environ["PYTHON_JULIACALL_HANDLE_SIGNAL"] = "no"

    opt = Config.get("optimization")
    if opt == "rapid":
        os.environ["PYTHON_JULIACALL_COMPILE"] = "min"
        os.environ["PYTHON_JULIACALL_OPTIMIZE"] = "0"
        os.environ["PYTHON_JULIACALL_MIN_OPTLEVEL"] = "0"
    elif opt == "latency":
        os.environ["PYTHON_JULIACALL_COMPILE"] = "yes"
        os.environ["PYTHON_JULIACALL_OPTIMIZE"] = "0"
        os.environ["PYTHON_JULIACALL_MIN_OPTLEVEL"] = "0"
    elif opt == "default":
        pass
    elif opt == "performance":
        os.environ["PYTHON_JULIACALL_COMPILE"] = "all"
        os.environ["PYTHON_JULIACALL_OPTIMIZE"] = "3"
        os.environ["PYTHON_JULIACALL_MIN_OPTLEVEL"] = "3"
    else:
        raise Exception(f"Unknown optimization setting '{opt}'")

    if sysimage.exists():
        logger.info("    Using custom sysimage: %s" % str(sysimage))
        os.environ["PYTHON_JULIACALL_SYSIMAGE"] = str(sysimage)
    else:
        logger.info("    Using default sysimage")
        logger.warning(
            " Consider executing `iesopt.create_sysimage()` once to create a custom sysimage that might offer improved"
            " performance"
        )

    logger.info("    Executable: %s" % juliapkg.executable())
    logger.info("    Project: %s" % juliapkg.project())

    os.environ["PYTHON_JULIACALL_BINDIR"] = str(Path(juliapkg.executable()).parent)

    import juliacall

    logger.info("Julia setup complete")

    custom_packages = list(Config.find("PKG_"))
    if len(custom_packages) > 0:
        logger.info("Installing custom Julia packages")

        try:
            juliacall.Main.seval("import Pkg")
        except Exception as e:
            logger.error(f"Failed to import Julia `Pkg`: {e}")

        for entry in custom_packages:
            name = entry[4:]
            try:
                juliacall.Pkg.add(name=name, version=Config.get(entry))
            except Exception as e:
                logger.error(f"Failed to install custom package '{name}': {e}")

    # Restoring potential SSL certificate.
    if _ssl is not None:
        logger.debug("Restoring local `SSL_CERT_FILE`")
        os.environ["SSL_CERT_FILE"] = _ssl

    return juliacall


def import_modules():
    logger.info("Importing Julia modules:")
    jl_import("IESopt")
    jl_import("JuMP")



import os
import importlib.metadata
from pathlib import Path

from ..util import logger, set_iesopt_module_attr
from .setup import setup_julia, import_modules


def initialize(target: Path, sysimage: Path):
    if os.getenv("IESOPT_DOCS_NOEXEC"):
        logger.warning("Detected docs environment (env. var. `IESOPT_DOCS_NOEXEC` is set), skip loading Julia")
        return

    logger.info("Integrated Energy System Optimization (IESopt)")
    logger.info("    © 2021 - now:  AIT Austrian Institute of Technology GmbH")
    logger.info("    Documentation: https://ait-energy.github.io/iesopt")

    juliacall = setup_julia(target, sysimage)
    julia = juliacall.Main
    set_iesopt_module_attr("julia", julia)
    set_iesopt_module_attr("juliacall", juliacall)

    import_modules()
    logger.info(
        "Loaded versions: py=%s, jl=%s" % (importlib.metadata.version("iesopt"), str(julia.pkgversion(julia.IESopt)))
    )

    set_iesopt_module_attr("IESopt", julia.IESopt)
    set_iesopt_module_attr("JuMP", julia.JuMP)

    set_iesopt_module_attr("jump_value", julia.seval("(x) -> JuMP.value.(x)"))
    set_iesopt_module_attr("jump_dual", julia.seval("(x) -> JuMP.dual.(x)"))
    set_iesopt_module_attr("jump_reduced_cost", julia.seval("(x) -> JuMP.reduced_cost.(x)"))
    set_iesopt_module_attr("jump_shadow_price", julia.seval("(x) -> JuMP.reduced_cost.(x)"))

    set_iesopt_module_attr("Docs.doc", julia.seval("(x) -> Docs.doc(x)"))

    return julia



import pytest
import re
import os
from unittest import mock


@mock.patch.dict(os.environ, {"IESOPT_DOCS_NOEXEC": "True"})
def test_julia_ci_version_consistent_with_python_package(project_root_path):
    import iesopt

    julia_python = iesopt.config.Config.DEFAULTS["IESOPT_JULIA"]

    # parse the CI action that sets the julia version
    julia_ci = "unknown"
    with open(f"{project_root_path}/.github/actions/set-versions/action.yml") as f:
        for line in f.readlines():
            if "VERSION_JULIA" in line:
                julia_ci = re.search('.*VERSION_JULIA=(.*)"', line).group(1)
                break

    assert julia_python == julia_ci, "CI Julia differs from Python Julia"


@pytest.fixture
def project_root_path(request):
    return request.config.rootpath
